"""Cartesian product of multiple named Spaces (also known as a dict of Spaces).

This Space produces samples which are dicts, where the values of those dicts
are drawn from the values of this Space.
"""
import gym.spaces

from akro.requires import requires_tf, requires_theano
from akro.space import Space


class Dict(gym.spaces.Dict, Space):
    """
    A dictionary of simpler spaces, e.g. Discrete, Box.

    Example usage:
        self.observation_space = spaces.Dict({"position": spaces.Discrete(2),
                                              "velocity": spaces.Discrete(3)})
    """

    @property  # pragma: no cover
    def flat_dim(self):
        """Return the length of the flattened vector of the space."""
        raise NotImplementedError

    def flatten(self, x):  # pragma: no cover
        """Return an observation of x with collapsed values.

        Args:
            x (:obj:`Iterable`): The object to flatten.

        Returns:
            Dict: A Dict where each value is collapsed into a single dimension.
                  Keys are unchanged.

        """
        raise NotImplementedError

    def unflatten(self, x):  # pragma: no cover
        """Return an unflattened observation x.

        Args:
            x (:obj:`Iterable`): The object to unflatten.

        Returns:
            np.ndarray: An array of x in the shape of self.shape.

        """
        raise NotImplementedError

    def flatten_n(self, xs):  # pragma: no cover
        """Return flattened observations xs.

        Args:
            xs (:obj:`Iterable`): The object to reshape and flatten

        Returns:
            np.ndarray: An array of xs in a shape inferred by the size of
                its first element.

        """
        raise NotImplementedError

    def unflatten_n(self, xs):  # pragma: no cover
        """Return unflattened observations xs.

        Args:
            xs (:obj:`Iterable`): The object to reshape and unflatten

        Returns:
            np.ndarray: An array of xs in a shape inferred by the size of
                its first element and self.shape.

        """
        raise NotImplementedError

    @requires_tf
    def to_tf_placeholder(self, name, batch_dims):
        """Create a tensor placeholder from the Space object.

        Args:
            name (str): name of the variable
            batch_dims (:obj:`list`): batch dimensions to add to the
                shape of the object.

        Returns:
            tf.Tensor: Tensor object with the same properties as
                the Dict where the shape is modified by batch_dims.

        """
        newdict = Dict()
        for key, space in self.spaces.items():
            newdict.spaces[key] = space.to_tf_placeholder(name, batch_dims)
        return newdict

    @requires_theano
    def to_theano_tensor(self, name, batch_dims):
        """Create a theano tensor from the Space object.

        Args:
            name (str): name of the variable
            batch_dims (:obj:`list`): batch dimensions to add to the
                shape of the object.

        Returns:
            theano.tensor.TensorVariable: Tensor object with the
                same properties as the Dict where the shape is
                modified by batch_dims.

        """
        newdict = Dict()
        for key, space in self.spaces.items():
            newdict.spaces[key] = space.to_theano_tensor(name, batch_dims)
        return newdict



import pickle
import unittest

from akro import Dict
from akro import Discrete
from akro import tf
from akro import theano
from akro.requires import requires_tf, requires_theano


class TestDict(unittest.TestCase):
    def test_pickleable(self):
        motion_dict = {'position': Discrete(2), 'velocity': Discrete(3)}
        sample = {
            'position': 1,
            'velocity': 2,
        }
        d = Dict(motion_dict)
        round_trip = pickle.loads(pickle.dumps(d))

        assert d.contains(sample)
        assert round_trip
        assert round_trip.contains(sample)

    def test_flat_dim(self):
        pass

    def test_flatten(self):
        pass

    def test_unflatten(self):
        pass

    def test_flatten_n(self):
        pass

    def test_unflatten_n(self):
        pass

    @requires_tf
    def test_convert_tf(self):
        d = Dict({'position': Discrete(2), 'velocity': Discrete(3)})
        tensor_dict = d.to_tf_placeholder('test', 1)
        assert isinstance(tensor_dict, Dict)
        assert all(
            [isinstance(c, tf.Tensor) for c in tensor_dict.spaces.values()])
        assert all([v.dtype == tf.int64 for v in tensor_dict.spaces.values()])

    @requires_theano
    def test_convert_theano(self):
        d = Dict({'position': Discrete(2), 'velocity': Discrete(3)})
        tensor_dict = d.to_theano_tensor('test', 1)
        assert isinstance(tensor_dict, Dict)
        assert all([
            isinstance(c, theano.tensor.TensorVariable)
            for c in tensor_dict.spaces.values()
        ])
        assert all(
            [space.dtype == 'int64' for space in tensor_dict.spaces.values()])



import unittest

import gym.spaces

import akro


class TestSpace(unittest.TestCase):
    def test_convert_box(self):
        obj = gym.spaces.Box(0.0, 1.0, (3, 4))
        box = akro.from_gym(obj)
        assert isinstance(box, akro.Box)
        assert (box.low == obj.low).all()
        assert (box.high == obj.high).all()
        assert box.shape == obj.shape

    def test_convert_dict(self):
        obj = gym.spaces.Dict({'foo': gym.spaces.Discrete(3)})
        dict = akro.from_gym(obj)
        assert isinstance(dict, akro.Dict)

    def test_convert_discrete(self):
        obj = gym.spaces.Discrete(3)
        disc = akro.from_gym(obj)
        assert isinstance(disc, akro.Discrete)

    def test_convert_tuple(self):
        obj = gym.spaces.Tuple((gym.spaces.Discrete(2),
                                gym.spaces.Discrete(3)))
        tup = akro.from_gym(obj)
        assert isinstance(tup, akro.Tuple)



"""Decorators used for calling tensorflow and theano functions safely."""

import functools

from akro import tf, theano


def requires_tf(func):
    """Check tf is installed before calling a function."""
    @functools.wraps(func)  # yapf: disable
    def check(*args, **kwargs):
        """Execute the function if the module is loaded.

        Args:
            args (array): positional args passed to the function.
            kwargs (dictionary): keyword args passed to the function.

        Returns:
            func: Result of the provided function.

        """
        if not tf:  # pragma: no cover
            raise RuntimeError(
                'Module "tensorflow" is required to use the function {}'.
                format(func))
        return func(*args, **kwargs)

    return check


def requires_theano(func):
    """Check theano is installed before calling a function."""
    @functools.wraps(func)  # yapf: disable
    def check(*args, **kwargs):
        """Execute the function if the module is loaded.

        Args:
            args (array): positional args passed to the function.
            kwargs (dictionary): keyword args passed to the function.

        Returns:
            func: Result of the provided function.

        """
        if not theano:  # pragma: no cover
            raise RuntimeError(
                'Module "theano" is required to use the function {}'.format(
                    func))
        return func(*args, **kwargs)

    return check
